//
// Copyright 2003 (c) Kevin Atkinson under the GNU GPL license version
// 2.0.  You should have received a copy of the GPL license along with
// this program if you did not you can find it at http://www.gnu.org/.
//

#include <windows.h>
#include <io.h>

#include <stdio.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdlib.h>
#include <assert.h>

#include <cmath>
#include <algorithm>

using std::swap;
using std::abs;
using std::min;
using std::max;

#include "avisynth_c.h"

////////////////////////////////////////////////////////////////////
//
// Logger
// 

static FILE * gclog = 0;
void alloc_clog();
void free_clog(void *, AVS_ScriptEnvironment *);

struct Log
{
  FILE * clog;
  FILE * flog;
  bool debug_p;
  int log_level;
  void use_clog();
  bool open_log_file(const char * fn);
  Log() : clog(0), flog(0), debug_p(false), log_level(2) {}
  void print(int l, const char * fmt, ...)
    __attribute__ ((format (printf, 3, 4)));
  ~Log() {if (flog) fclose(flog);}
};

////////////////////////////////////////////////////////////////////
//
// Utility Structures
// 

typedef unsigned char byte;

static inline bool is_odd(int x) {return x & 1;}
static inline bool is_even(int x) {return !is_odd(x);}

static inline unsigned floor2(unsigned x, unsigned mul) {return x & ~(mul - 1);}
static inline unsigned ceil2(unsigned x, unsigned mul) {return floor2(x + mul-1, mul);}

#define REGPARM __attribute__ (( regparm(3) ))

class VideoFramePtr
{
  AVS_VideoFrame * ptr;
  VideoFramePtr(const VideoFramePtr &);
  void operator= (const VideoFramePtr &);
public:
  VideoFramePtr(AVS_VideoFrame * p = 0) : ptr(p) {}
  void operator= (AVS_VideoFrame * p) {assert(ptr == 0); ptr = p;}
  operator AVS_VideoFrame * () {return ptr;}
  void reset(AVS_VideoFrame * p = 0) {
    if (ptr) avs_release_video_frame(ptr); ptr = p;}
  AVS_VideoFrame * release() {AVS_VideoFrame * tmp = ptr; ptr = 0; return tmp;}
  ~VideoFramePtr() {if (ptr) avs_release_video_frame(ptr);}
};

template <typename T, unsigned S>
struct Cache
{
  typedef T Type;
  static const unsigned size = 1 << S;
  static const unsigned mask = size - 1;
  T data[size];
  T & operator[] (int num) {return data[num & mask];}
};

static const float DELTA = 0.001;

////////////////////////////////////////////////////////////////////
//
// Core Datatypes
//

static const int HSCALE = 2;

struct CompareParms;
struct CompareData;
struct MutableRawFrame; 
struct ConstRawFrame;
struct BasicFrameInfo;

enum CPUType {CPU_GEN= 0, CPU_MMX = 1, CPU_ISSE = 2, CPU_SSE = 3, CPU_SSE2 = 4};
CPUType CPU_TYPE = CPU_GEN;
static const int NUM_CPU_TYPES = 5;

typedef float CompareFields(const CompareParms * p, 
			    const BasicFrameInfo *, const BasicFrameInfo *);
extern "C" CompareFields compare_fields_gen, compare_fields_sse;

// Note: *_isse functions below (compare, subsample) do NOT call emms

typedef REGPARM void (SubsampleLine)(const byte *, short *, size_t);
extern "C" SubsampleLine subsample_line_yuy2_2_gen, subsample_line_yv12_2_gen,
                         subsample_line_yuy2_2_isse, subsample_line_yv12_2_isse,
                         subsample_line_yuy2_2_isse2, subsample_line_yv12_2_isse2;

typedef REGPARM void CompareFieldsRow(CompareData & d, short *);
extern "C" CompareFieldsRow compare_fields_row_gen, compare_fields_row_isse;

// these are used by assembly code, do not call from C++ code
typedef void AsmFunc(...);
extern "C" AsmFunc asm_compare_fields_row_loop_isse, asm_compare_fields_row_loop_isse2;
extern "C" AsmFunc asm_compare_fields_mul_sse, asm_compare_fields_mul_sse2;

const struct CPUInfo
{
  CPUType type;
  const char * str;
  CompareFields * compare_fields;
  SubsampleLine * subsample_line[2];
  int align;
  int fwidth[2];
  CompareFieldsRow * compare_fields_row;
  AsmFunc * asm_compare_fields_row_loop;
  AsmFunc * asm_compare_fields_mul;
} cpu_info[NUM_CPU_TYPES] = {
  {CPU_GEN, "Generic", compare_fields_gen, 
   {subsample_line_yuy2_2_gen, subsample_line_yv12_2_gen}, 1, {4, 2}, 
   compare_fields_row_gen, 0, 0},
  {CPU_GEN, "Generic", compare_fields_gen, 
   {subsample_line_yuy2_2_gen, subsample_line_yv12_2_gen}, 1, {4, 2}, 
   compare_fields_row_gen, 0, 0},
  {CPU_ISSE, "Integer SSE", compare_fields_gen,
   {subsample_line_yuy2_2_isse, subsample_line_yv12_2_isse}, 8, {16, 8}, 
   compare_fields_row_isse, 
   asm_compare_fields_row_loop_isse, 0},
  {CPU_SSE, "SSE", compare_fields_sse,
   {subsample_line_yuy2_2_isse, subsample_line_yv12_2_isse}, 8, {16, 8},
   compare_fields_row_isse, 
   asm_compare_fields_row_loop_isse, asm_compare_fields_mul_sse},
  {CPU_SSE, "SSE2", compare_fields_sse,
   {subsample_line_yuy2_2_isse2, subsample_line_yv12_2_isse2}, 16, {32, 16},
   compare_fields_row_isse, 
   asm_compare_fields_row_loop_isse2, asm_compare_fields_mul_sse2}
};

enum VideoFormat {VF_YUY2, VF_YV12};
int pixel_width_p2[2] = {1,0};

enum FieldOrder {FO_BFF, FO_TFF, FO_UNKNOWN};
struct Filter;
struct FieldDiff
{
  int num;
  float diff;
  int   what;
  AVS_VideoFrame * frame;
  void clear() {num = -9; diff = 1.0; what = 2;}
  FieldDiff() {clear();}
  FieldDiff(int n, float d, int w) : num(n), diff(d), what(w) {}
};

struct BasicFrameInfo {
  const byte * data; // real data start is data + pitch
  const byte * stop; // real data start is data + pitch
  int pitch;
  SubsampleLine * subsample_line;
  BasicFrameInfo() : data(0) {}
};

struct FrameInfo : public BasicFrameInfo
{
  int num;
  VideoFramePtr frame;
  int parity;
  int cpu;
  int begin_ignore;
  int end_ignore;
  void release() {frame.reset();}
  void clear() {frame.reset(); num = -9;}
  FrameInfo() : num(-9), frame(0) {}
  ~FrameInfo() {if (frame) avs_release_video_frame(frame);}
};

struct RenderInfo 
{
  int num;
  int parity;
  int even;
  int odd; // only used for weave
  RenderInfo() : num(-9) {}
  void clear() {num = -9;}
  void bob(int fn) {even = odd = fn;}
  void weave(int fn, int e, int o) {even = fn + e; odd = fn + o;}
};

// hand written assembler code uses this structure
// do not modify with out adjusting the code in tel-f.s
struct CompareParms
{
  int cpu;
  CompareFieldsRow * compare_fields_row;
  AsmFunc * cf_row_loop;
  AsmFunc * cf_mul;
  int row_size;
  int s_height;
  int s_width;
  int s_pitch;
  void * data;
  float scale;
  Log * log;
  short * a, * b, * c, * d, * e;
  short * prev0, * cur0, * next0;
  float * prev1, * cur1, * next1;
  CompareParms() : data(0) {}
  void alloc_data();
};

struct Analysis
{
  int MAX_LAST_SET;

  float T1_DEF, T2_DEF;
  float T1_MAX, T2_MAX;

  float FLOOR, FLOOR_ADJ;
  float SML_PEAK, SML_TAIL_DIFF;
  float LRG_PEAK, LRG_TAIL_DIFF;

  float pure_tel;
  float old_thres1;
  float old_thres2;
  float thres1;
  float thres2;
  float vals[3];
  int last_set;
  int ready;
  int num; // frame number the next adv will give
  
  Log * log;

  void reset(int num);
  FieldDiff adv(float val); // delay of 2
};

struct Filter
{
  Log log;
  
  float pure_tel;
  
  FieldOrder field_order;
  CPUType cpu;
  VideoFormat video_format;

  int ratio_n;
  int denm;
  int round_div_n;

  size_t width;
  size_t row_size;
  size_t height;

  AVS_Clip * true_child;
  AVS_Clip * child;
  AVS_Clip * bob_src;
  AVS_Clip * weave_src;
  AVS_ScriptEnvironment * env;
  AVS_VideoInfo * dest_vi;

  bool unaligned;

  int prev_src_fn;
  int cur_src_fn;
  int prev_fn;
  int cur_fn;
  Analysis anal;
  CompareParms comp;
  
  int accum_n;
  int num_src_frames;
  
  Filter() : true_child(0), child(0), bob_src(0), weave_src(0), prev_fn(-9) {}
  ~Filter() {if (bob_src) avs_release_clip(bob_src); 
             if (weave_src) avs_release_clip(weave_src); 
	     if (child && true_child != child) avs_release_clip(child);
             if (comp.data) free(comp.data);}
  
  //
  // Cache
  //
  
  Cache<FieldDiff,4>  fd_cache;
  Cache<RenderInfo,6> render_cache;
  Cache<FrameInfo,1>  src_cache;

  int get_src_parity(int fn) {return field_order ^ (fn & 1);}
  REGPARM void get_src_frame(FrameInfo &, int fn);
  
  int diff(int d1, int d2);
  bool same_group(int x, int y) {return diff(x,y) <= 0;}
  bool similar_group(int x, int y) {return diff(x,y) <= 1;}
  bool diff_group(int x, int y) {return diff(x,y) == 2;}
  
  void clear_cache();
  void adv_fd_cache_to(int fn);

  //
  //
  //

  
  float compare(int num);

  void reset();
  bool chose_next_frame();
  void get_frame_info();
  AVS_VideoFrame * render(const RenderInfo &);

  void ready_frame(int num);
  
  static AVS_VideoFrame * get_frame(AVS_FilterInfo *, int num);
  static int get_parity(AVS_FilterInfo *, int num);
  static void free_filter(AVS_FilterInfo *);
};

/////////////////////////////////////////////////////////////////////
//
// Initilization Code
//

#define ERR(str) do {ret = avs_new_value_error(str); goto exit;} while (0)
#define CHECK_RANGE(str,var,low,high) do {\
    if (var < low || var > high) ERR(#str " must be in the range "\
                                     #low " - " #high "."); } while(0)
#define CHECK_GT(str, var, low)  do {\
    if (var < low) ERR(str " must be greater than  " #low "."); } while(0)
#define CHECK_ORDER(str1, str2, var1, var2) do {\
    if (var1 > var2) ERR(str1 " must be less than " str2 ".");}while(0)
#define CHECK_ORDER_ANAL(var1, var2) CHECK_ORDER(#var1, #var2, f->anal.var1, f->anal.var2)
#define SET_I_ANAL(what, def) do {\
    tmp = avs_array_elt(args, i++);\
    if (avs_defined(tmp)) f->anal.what = avs_as_int(tmp);\
    else f->anal.what = def;} while (0)
#define SET_F_ANAL(what, def) do {\
    tmp = avs_array_elt(args, i++);\
    if (avs_defined(tmp)) f->anal.what = avs_as_float(tmp);\
    else f->anal.what = def;} while (0)
#define SET_F_R_ANAL(what, def) do {SET_F_ANAL(what, def); \
                               CHECK_RANGE(#what, f->anal.what, 0.0, 1.0); } while (0)
#define SET_F_0_ANAL(what, def) do {SET_F_ANAL(what, def); \
                               CHECK_GT(#what, f->anal.what, 0.0); } while (0)
#define SET_F_1_ANAL(what, def) do {SET_F_ANAL(what, def); \
                               CHECK_GT(#what, f->anal.what, 1.0); } while (0)

AVS_Value smart_decimate(AVS_ScriptEnvironment * env, AVS_Value args, void * )
{
  AVS_Value ret;
  AVS_Value tmp,tmp1;
  AVS_FilterInfo * fi;
  int numr, denm;
  int i = 1; 
  AVS_Clip * clip = avs_new_c_filter(env, &fi, args.d.array[0], 1);
  Filter * f = new Filter;
  f->true_child = fi->child;
  f->env        = fi->env;
  f->dest_vi    = &fi->vi;
  f->anal.log = &f->log;
  f->comp.log    = &f->log;
  fi->user_data = f;
  fi->get_frame   = Filter::get_frame;
  fi->get_parity  = Filter::get_parity;
  fi->free_filter = Filter::free_filter;
  AVS_Value sclip = avs_new_value_clip(fi->child);
  AVS_Value wargs = avs_new_value_array(&sclip, 1);
  {
    if (avs_is_yuy2(&fi->vi))
      f->video_format = VF_YUY2;
    else if (avs_is_yv12(&fi->vi))
      f->video_format = VF_YV12;
    else
      ERR("Video must be YUY2 or YV12");

    if (!avs_is_field_based(&fi->vi)) {
      AVS_Value rclip = avs_invoke(env, "SeparateFields", wargs, 0); 
      if (avs_is_error(rclip)) ERR(avs_as_error(rclip));
      f->child = avs_take_clip(rclip, env);
      avs_release_value(rclip);
    } else {
      f->child = f->true_child;
    }

    const AVS_VideoInfo * child_vi = avs_get_video_info(f->child);
    if (!avs_is_field_based(child_vi)) ERR("Video is not interlaced or field based.");

    if (child_vi->width < 32) ERR("The width must be at least 32.");
    if (child_vi->height < 16) ERR("The height must be at least 32.");
    
    if (child_vi->width % 2 != 0) ERR("The width must be a multiple of two.");

    tmp = avs_array_elt(args, i++);
    tmp1 = avs_array_elt(args, i++);
    if (avs_defined(tmp) && avs_defined(tmp1)) {
      numr = avs_as_int(tmp);
      denm = avs_as_int(tmp1);
      CHECK_GT("numr", numr, 0);
      CHECK_GT("denm", denm, 0);
      if (numr > denm)
	ERR("Decimation ratio can not be larger than 1.0.");
    } else if (avs_defined(tmp) || avs_defined(tmp1)) {
      ERR("Both numr and denm must either be set or unset.");
    } else {
      numr = 24;
      denm = 60;
    }
    {
      unsigned x=numr, y=denm;
      while (y) { // find gcd
	unsigned t = x%y; x = y; y = t;
      }
      f->denm = numr * 2 / x;
      f->ratio_n = denm * 2 / x;
    }
   
    if (f->ratio_n/f->denm > 12)
      ERR("Decimation ratio is less than 1/12");

    tmp = avs_array_elt(args, i++);  // bob
    tmp1 = avs_array_elt(args, i++); // weave

    const AVS_VideoInfo * weave_vi = 0;

    if (avs_defined(tmp1)) {
      f->weave_src = avs_take_clip(tmp1, f->env);
      weave_vi = avs_get_video_info(f->weave_src);
      if (abs((float)child_vi->fps_numerator/child_vi->fps_denominator -
	      (float)weave_vi->fps_numerator/weave_vi->fps_denominator) > 0.001)
	ERR("Double Weave source frame rate incorrect.");
      avs_release_value(tmp1);
    } else {
      AVS_Value rclip = avs_invoke(env, "DoubleWeave", wargs, 0);
      if (avs_is_error(rclip)) ERR(avs_as_error(rclip));
      f->weave_src = avs_take_clip(rclip, env);
      avs_release_value(rclip);
      weave_vi = avs_get_video_info(f->weave_src);
    }

    *f->dest_vi = *weave_vi;

    if (avs_defined(tmp)) {
      f->bob_src = avs_take_clip(tmp, f->env);
      const AVS_VideoInfo * weave_vi = avs_get_video_info(f->weave_src);
      const AVS_VideoInfo * bob_vi = avs_get_video_info(f->bob_src);
      if (weave_vi->width != bob_vi->width || weave_vi->height != bob_vi->height)
	ERR("Bob source video has incorrect dimmensions.");
      if (abs((float)weave_vi->fps_numerator/weave_vi->fps_denominator -
	      (float)bob_vi->fps_numerator/bob_vi->fps_denominator) > 0.001)
	ERR("Bob source frame rate incorrect.");
      avs_release_value(tmp);
    } else {
      AVS_Value rclip = avs_invoke(env, "Bob", wargs, 0);
      if (avs_is_error(rclip)) ERR(avs_as_error(rclip));
      f->bob_src = avs_take_clip(rclip, env);
      avs_release_value(rclip);
    }

    tmp = avs_array_elt(args, i++); // tel
    if (avs_defined(tmp))  f->pure_tel = avs_as_float(tmp);
    else                   f->pure_tel = 0.50;    
    f->anal.pure_tel = f->pure_tel;
    CHECK_RANGE("tel", f->pure_tel, 0.0, 1.0);
    
    tmp  = avs_array_elt(args, i++); // noise
    tmp1 = avs_array_elt(args, i++); // t_max
    if (avs_defined(tmp1)) {
      f->anal.T1_MAX = avs_as_float(tmp1);
    } else if (avs_defined(tmp)) {
      f->anal.T1_MAX = expf(17.65*avs_as_float(tmp) - 20.62);
    } else {
      f->anal.T1_MAX = expf(17.65*0.50 - 20.62);
    }
    f->anal.T2_MAX = 3*f->anal.T1_MAX;

    tmp  = avs_array_elt(args, i++); // unaligned
    f->unaligned = avs_defined(tmp) ? avs_as_bool(tmp) : false; 
  
    tmp  = avs_array_elt(args, i++); // max_last_set
    if (avs_defined(tmp))
      f->anal.MAX_LAST_SET = avs_as_int(tmp);
    else
      f->anal.MAX_LAST_SET = (5*denm)/numr + 1;
    CHECK_GT("max_last_set", f->anal.MAX_LAST_SET, 0);
 
    SET_F_R_ANAL(T1_DEF, 5e-12);
    SET_F_R_ANAL(T2_DEF, 9e-12);
  
    tmp  = avs_array_elt(args, i++); // t1_max
    if (avs_defined(tmp)) f->anal.T1_MAX = avs_as_float(tmp);
    tmp  = avs_array_elt(args, i++); // t2_max
    if (avs_defined(tmp)) f->anal.T2_MAX = avs_as_float(tmp);

    CHECK_RANGE("t1_max", f->anal.T1_MAX, 0.0, 1.0);  
    CHECK_RANGE("t2_max", f->anal.T2_MAX, 0.0, 1.0);  
    CHECK_ORDER_ANAL(T1_DEF, T2_DEF);
    CHECK_ORDER_ANAL(T1_MAX, T2_MAX);

    SET_F_R_ANAL(FLOOR, 0.05);
    SET_F_0_ANAL(FLOOR_ADJ, 0.50);
    SET_F_1_ANAL(SML_PEAK, 1.1);
    SET_F_0_ANAL(SML_TAIL_DIFF, 0.35);
    SET_F_1_ANAL(LRG_PEAK, 2.0);
    SET_F_0_ANAL(LRG_TAIL_DIFF, 0.15);
    CHECK_ORDER_ANAL(SML_PEAK, LRG_PEAK);
  
    tmp = avs_array_elt(args, i++);
    if (avs_defined(tmp)) {
      f->cpu = (CPUType)avs_as_int(tmp);
      CHECK_RANGE("cpu", f->cpu, 0, 4);
    } else {
      f->cpu = CPU_TYPE;
    }
  
    tmp = avs_array_elt(args, i++);
    if (avs_defined(tmp)) f->log.log_level = avs_as_int(tmp);
    CHECK_RANGE("log_level", f->log.log_level, 0, 5);
  
    tmp = avs_array_elt(args, i++);
    if (avs_defined(tmp)) f->log.open_log_file(avs_as_string(tmp));
  
    tmp = avs_array_elt(args, i++);
    if (avs_defined(tmp) && avs_as_bool(tmp)) f->log.use_clog();

    tmp = avs_array_elt(args, i++);
    if (avs_defined(tmp) && avs_as_bool(tmp)) f->log.debug_p = true;

    f->log.print(0, "CPU Type is %s", cpu_info[f->cpu].str);
  
    f->num_src_frames = weave_vi->num_frames;

    f->width = weave_vi->width;
  
    FrameInfo frame;
    f->get_src_frame(frame, 0);

    f->row_size = avs_get_row_size(frame.frame);  
    f->height   = avs_get_height(frame.frame);
  
    if (frame.cpu != f->cpu)
      f->log.print(0, "WARNING: Using %s optimizations instead of %s when reading frame data due to unaligned data.",
		   cpu_info[frame.cpu].str, cpu_info[f->cpu].str);
    if (frame.begin_ignore != 0)
      f->log.print(0, "WARNING: First %d pixels in each row will be ignored.",
		   frame.begin_ignore);
    if (frame.end_ignore != 0)
      f->log.print(0, "WARNING: Last %d pixels in each row will be ignored.",
		   frame.end_ignore);

    if (avs_get_parity(fi->child, 0)) {
      f->field_order = FO_TFF;
      f->log.print(0, "Field Order Is TFF");
    } else {
      f->field_order = FO_BFF;
      f->log.print(0, "Field Order Is BFF");
    }
  
    f->log.print(3, "t1_max = %#g  t2_max = %#g  max_last_set = %d",
		 f->anal.T1_MAX, f->anal.T2_MAX, f->anal.MAX_LAST_SET);
  
    // Set compare parms

    f->comp.cpu = f->cpu;
    f->comp.compare_fields_row = cpu_info[f->cpu].compare_fields_row;
    f->comp.cf_row_loop = cpu_info[f->cpu].asm_compare_fields_row_loop;
    f->comp.cf_mul      = cpu_info[f->cpu].asm_compare_fields_mul;

    f->comp.row_size = f->row_size;
    f->comp.s_height = f->height * 2 / 3;
    f->comp.s_width  = f->width/HSCALE;

    f->comp.alloc_data();

    f->comp.s_width  = (f->width - frame.begin_ignore - frame.end_ignore)/HSCALE;

    f->comp.scale  = 256 * 6 * (2*HSCALE);
    f->comp.scale *= f->comp.scale;
    f->comp.scale *= f->comp.scale;
    f->comp.scale *= f->comp.s_width * f->height / 1.5f;
    f->comp.scale  = 1/f->comp.scale;

    avs_clear_property(f->dest_vi, AVS_IT_BFF);
    avs_clear_property(f->dest_vi, AVS_IT_TFF);
    avs_clear_property(f->dest_vi, AVS_IT_FIELDBASED);
    f->dest_vi->num_frames = (int)(f->dest_vi->num_frames*f->denm/f->ratio_n);
    avs_set_fps(f->dest_vi, f->dest_vi->fps_numerator * numr, 
		f->dest_vi->fps_denominator * denm);

    if (log2((f->dest_vi->num_frames + 1.5) * (double)f->denm) > 31.0)
      ERR("log2(num_frames*numr*4) > 31");
  }
  ret = avs_new_value_clip(clip);
 exit:
  avs_release_value(wargs);
  avs_release_value(sclip);
  avs_release_clip(clip); clip = 0;
  return ret;
}

const char * avisynth_c_plugin_init(AVS_ScriptEnvironment * env)
{
  avs_add_function(
    env, "SmartDecimate", 
    "c[numr]i[denm]i[bob]c[weave]c" 
    "[tel]f[noise]f[t_max]f"
    "[unaligned]b"
    "[max_last_set]i" "[t1_def]f[t2_def]f[t1_max]f[t2_max]f"
    "[floor]f[floor_adj]f"
    "[sml_peak]f[sml_tail_diff]f[lrg_peak]f[lrg_tail_diff]f"
    "[cpu]i"
    "[log_level]i[log_file]s[console]b[debug_print]b",
    smart_decimate, 0);
  avs_at_exit(env, free_clog, 0);
  
  // set up cpu type
  long cpuf = avs_get_cpu_flags(env);
  if (cpuf & AVS_CPU_SSE2)
    CPU_TYPE = CPU_SSE2;
  else if (cpuf & AVS_CPU_SSE)
    CPU_TYPE = CPU_SSE;
  else if (cpuf & AVS_CPU_INTEGER_SSE)
    CPU_TYPE = CPU_ISSE;
  else
    CPU_TYPE = CPU_GEN;
  return "";
}

//////////////////////////////////////////////////////////////////////
//
// Filter class implementation
//

AVS_VideoFrame * Filter::get_frame(AVS_FilterInfo * fi, int num)
{
  Filter * f = (Filter *)fi->user_data;
  const RenderInfo & rinf = f->render_cache[num];
  if (rinf.num != num) 
    f->ready_frame(num);
  return f->render(rinf);
}

int Filter::get_parity(AVS_FilterInfo * fi, int num)
{
  Filter * f = (Filter *)fi->user_data;
  const RenderInfo & rinf = f->render_cache[num];
  if (rinf.num != num) 
    f->ready_frame(num);
  return rinf.parity;
}

void Filter::free_filter(AVS_FilterInfo * fi)
{
  Filter * f = (Filter *)fi->user_data;
  delete f;
  fi->user_data = 0;
}

void Filter::ready_frame(int num)
{
  assert(num >= 0);
  assert(num < dest_vi->num_frames);
  cur_fn = num;
  if (prev_fn + 1 == cur_fn) {
    chose_next_frame();
  } else if (prev_fn >= 0 && prev_fn <= num && prev_fn + 8 >= num) {
    for (int i = prev_fn + 1; i < num; ++i)
      ready_frame(i);
    ready_frame(num);
    return;
  } else {
    reset();
  }
  get_frame_info();
  prev_fn = cur_fn;
}

void Filter::reset()
{
  log.print(1, "Reseting to %d", cur_fn);
  clear_cache();
  int ideal_n = cur_fn*ratio_n;
  cur_src_fn = ideal_n/denm;
  accum_n = ideal_n - cur_src_fn*denm;
  round_div_n = denm;
  anal.reset(cur_src_fn);
  anal.adv(compare(anal.num + 2));
  anal.adv(compare(anal.num + 2));
}

float Filter::compare(int num)
{
  if (num < 0 || num >= num_src_frames)
    return 1.0;
  FrameInfo & cur  = src_cache[num];
  FrameInfo & next = src_cache[num+1];
  if (cur.num != num)      get_src_frame(cur, num);
  if (next.num != num + 1) get_src_frame(next, num+1);
  assert(cur.parity != next.parity);
  float res;
  if (cur.parity)
    res = cpu_info[cpu].compare_fields(&comp, &cur, &next);
  else
    res = cpu_info[cpu].compare_fields(&comp, &next, &cur);
  cur.clear();
  return res;
}

void Filter::get_src_frame(FrameInfo & d, int fn)
{
  d.num = fn;
  if (d.frame) d.release();
  d.frame = avs_get_frame(child, fn);
  d.parity = avs_get_parity(child, fn);
  d.data = avs_get_read_ptr(d.frame);
  d.pitch = avs_get_pitch(d.frame);
  unsigned int align;
  d.cpu = cpu + 1;
 loop: {
    d.cpu--;
    assert(d.cpu >= 0);
    align = cpu_info[d.cpu].align;
    unsigned int pitch_align = d.pitch & (align - 1);
    if (pitch_align != 0) goto loop;
    d.subsample_line = cpu_info[d.cpu].subsample_line[video_format];
    unsigned long data_i = (unsigned long)d.data;
    d.begin_ignore = ceil2(data_i, align) - data_i;
    unsigned int row_size = avs_get_row_size(d.frame);
    unsigned int fwidth = cpu_info[d.cpu].fwidth[video_format];
    unsigned int rsize = floor2(row_size - d.begin_ignore, fwidth);
    d.end_ignore = row_size - rsize - d.begin_ignore;
    d.begin_ignore >>= pixel_width_p2[video_format];
    d.end_ignore   >>= pixel_width_p2[video_format];
    d.stop = d.data + d.pitch*height;
    d.data -= d.pitch;
    if (unaligned && (d.begin_ignore || d.end_ignore)) goto loop;
  }
}

void Filter::adv_fd_cache_to(int fn)
{
  assert(anal.num <= fn);
  while (anal.num <= fn)
  {
    FieldDiff fd = anal.adv(compare(anal.num + 2));
    fd_cache[fd.num] = fd;
  }
}
  
int Filter::diff(int d1, int d2) {
  if (d2 < d1) swap(d1,d2);
  d1 += cur_src_fn; d2 += cur_src_fn;
  int mx = 0;
  for (; d1 < d2; ++d1) {
    FieldDiff & d = fd_cache[d1];
    if (d.num != d1) adv_fd_cache_to(d1);
    mx = max(mx, d.what);
  }
  return mx;
}

void Filter::clear_cache() 
{
  for (unsigned i = 0; i != fd_cache.size; ++i)
    fd_cache.data[i].clear();
  for (unsigned i = 0; i != render_cache.size; ++i)
    render_cache.data[i].clear();
}

//////////////////////////////////////////////////////////////////////
//
// Compare
//

void CompareParms::alloc_data()
{
  s_pitch = s_width;
  if (s_pitch % 16 != 0) s_pitch += 16 - (s_pitch % 16);
  int data_size = s_pitch*(sizeof(short)*8 + sizeof(float)*3) + 32;
  data = malloc(data_size);
  memset(data, 0, data_size);
  unsigned long addr = (unsigned long)data;
  if (addr % 32 != 0) addr += 32 - (addr % 32);
  short * sp = (short *)addr;
  a = sp + 0*s_pitch;
  b = sp + 1*s_pitch;
  c = sp + 2*s_pitch;
  d = sp + 3*s_pitch;
  e = sp + 4*s_pitch;
  prev0 = sp + 5*s_pitch;
  cur0  = sp + 6*s_pitch;
  next0 = sp + 7*s_pitch;
  float * fp = (float *)(sp + 8*s_pitch);
  prev1 = fp + 0*s_pitch;
  cur1  = fp + 1*s_pitch;
  next1 = fp + 2*s_pitch;
}

struct CompareData : public CompareParms {
  CompareData(const CompareParms & d) : CompareParms(d) {}
  BasicFrameInfo s_l[2];
  int line;
};

extern "C" REGPARM void memclear(void * start, void * stop);

extern "C" REGPARM void subsample_line(CompareData & D, short * o, int i);

#define EMMS asm("emms" : : : "st", "st(1)", "st(2)", "st(3)",\
                              "st(4)", "st(5)", "st(6)", "st(7)")

float compare_fields_gen(const CompareParms * p0, 
			 const BasicFrameInfo * d0, const BasicFrameInfo * d1)
{
  CompareData D(*p0);
  D.s_l[0] = *d0; D.s_l[1] = *d1;
  D.line = 0;
  float accum = 0;
  memclear(D.prev0, D.prev0 + D.s_pitch);
  memclear(D.prev1, D.prev1 + D.s_pitch);
  subsample_line(D, D.d, 0);
  subsample_line(D, D.e, 1);
  if (D.cpu != CPU_GEN) EMMS;
  do
  {
    D.compare_fields_row(D,D.cur0);
    D.compare_fields_row(D,D.next0);
    if (D.cpu != CPU_GEN) EMMS;
    for (int x = 0; x < D.s_width; ++x)
      D.cur1[x] = (D.prev0[x]+D.next0[x])*(float)D.cur0[x];
    swap(D.prev0,D.next0);
    D.compare_fields_row(D,D.cur0);
    D.compare_fields_row(D,D.next0);
    if (D.cpu != CPU_GEN) EMMS;
    for (int x = 0; x < D.s_width; ++x)
      D.next1[x] = (D.prev0[x]+D.next0[x])*(float)D.cur0[x];
    swap(D.prev0,D.next0);
    for (int x = 0; x < D.s_width; ++x)
      accum += (D.prev1[x]+D.next1[x])*D.cur1[x];
    swap(D.prev1,D.next1);
  } while (D.s_l[0].data < D.s_l[0].stop || D.s_l[1].data < D.s_l[1].stop);
  accum *= D.scale;
  return accum;
}

void compare_fields_row_gen(CompareData & D, short * o)
{
  swap(D.a,D.d);
  swap(D.b,D.e);
  int i = D.line & 1; // 0 if even, 1 if odd
  subsample_line(D, D.c, i);
  subsample_line(D, D.e, i);
  i = i ^ 1; // 1 if even, 0 if odd
  subsample_line(D, D.d, i);
  o += D.s_width;
  // now compare lines
  short * a = D.a + D.s_width, * b = D.b + D.s_width, 
        * c = D.c + D.s_width, * d = D.d + D.s_width,
        * e = D.e + D.s_width;
  for (int x = -D.s_width; x != 0; ++x) 
  {
    o[x] = abs((a[x] + 4*c[x] + e[x]) - 3*(b[x] + d[x])) + 1;
  }
  ++D.line;
}

void subsample_line(CompareData & D, short * o, int i)
{
  BasicFrameInfo & d = D.s_l[i];
  d.data += d.pitch;
  if (d.data < d.stop) {
    d.subsample_line(d.data, o, D.row_size);
  } else {
    memclear(o, o + D.s_width);
  }
}

// These are now implemented in assembly.  They are left here for
// documentation and testing purposes.

#if 0

void subsample_line_yuy2_2_gen(const byte * x, short * o, size_t row_size)
{
  row_size = (row_size / 4) * 4;
  const byte * stop = x + (row_size / 8) * 8;
  const byte * stop2 = x + row_size;
  short a,b;
  a = 0;
  while (x < stop) {
    b = x[2];
    o[0] = a + x[0]*2 + b;
    a = x[6];
    o[1] = b + x[4]*2 + a;
    x += 8;
    o += 2;
  }
  if (x < stop2) {
    b = x[2];
    o[0] = a + x[0] + b;
  }
}

void subsample_line_yv12_2_gen(const byte * x, short * o, size_t row_size)
{
  row_size = (row_size / 2) * 2;
  const byte * stop = x + (row_size / 4) * 4;
  const byte * stop2 = x + row_size;
  short a,b;
  a = 0;
  while (x < stop) {
    b = x[1];
    o[0] = a + x[0]*2 + b;
    a = x[3];
    o[1] = b + x[2]*2 + a;
    x += 4;
    o += 2;
  }
  if (x < stop2) {
    b = x[1];
    o[0] = a + x[0] + b;
  }
}

#endif

//////////////////////////////////////////////////////////////////////
//
// Analysis
//

static const float T_MIN_RATIO = 1.1;

static const float SML_T_MIN = 0.50, SML_T_MAX = 0.8;
static const float LRG_T1_MIN = 0.01, LRG_T1_MAX = 0.30;
static const float LRG_T2_MIN = 0.60, LRG_T2_MAX = 0.90;

void Analysis::reset(int n)
{
  thres1 = 0;
  thres2 = 0;
  thres1 = T1_DEF;
  thres2 = T2_DEF;
  last_set = -1;
  num = n - 2 - 2;
  ready = 0;
}

static inline float w_avg(float x, float y, float w)
{
  return x*(1-w) + y*w;
}
static inline float clip(float x, float min, float max)
{
  if (x < min) return min;
  if (x > max) return max;
  return x;
}
static inline float clip(float x, float min, float max, float r_min, float r_max)
{
  return clip(x, w_avg(min,max,r_min), w_avg(min,max,r_max));
}

FieldDiff Analysis::adv(float val)
{
  ++num;
  if (ready < 2) {

    vals[ready] = val;
    ready++;
    return FieldDiff();

  } else {

    bool has_peak = false;

    vals[2] = val;
    if (vals[0] < vals[1] && vals[2] < vals[1]) 
    {
      float p = vals[1] / vals[0];
      float n = vals[1] / vals[2];
      float peak = min(p,n);
      float max_val = max(vals[0], vals[2]);
      float tail_max = max(vals[0],vals[2]);
      float tail_min = min(vals[0],vals[2]);
      float rel_tail_diff = (tail_max - tail_min) / (vals[1] - tail_min);
      float self_tail_diff = 2.0f * (tail_max - tail_min) / (tail_max + tail_min);
      float tail_diff = sqrtf(rel_tail_diff * self_tail_diff);
      log->print(4, "%d:  Peak %0.3f  Tail Diff %0.3f  (%0.3f %0.3f)", 
                 num, peak, tail_diff, rel_tail_diff, self_tail_diff); 
      if (max_val < T2_MAX && peak > LRG_PEAK && tail_diff < LRG_TAIL_DIFF)
      {
        log->print(4, " is T2");
        last_set = 0;
        thres1 = clip(thres1, tail_max, vals[1], LRG_T1_MIN, LRG_T1_MAX);
        thres2 = clip(thres2, tail_max, vals[1], LRG_T2_MIN, LRG_T2_MAX);
        has_peak = true;
      }
      else if (max_val < T2_MAX && peak > SML_PEAK && tail_diff < SML_TAIL_DIFF)
      {
        last_set = 0;
        if (thres2 > vals[1] && (pure_tel > 0.45f || thres1 >= tail_min)) {
          log->print(4, " is T1-T1");
          thres1 = clip(thres1, tail_max, vals[1], SML_T_MIN, SML_T_MAX);
          thres2 = max(thres2, thres1*T_MIN_RATIO);
        } else {
          log->print(4, " is T1-T2");
          thres2 = clip(thres2, tail_max, vals[1], SML_T_MIN, SML_T_MAX);
          thres1 = min(thres1, thres2/T_MIN_RATIO);
        }
        has_peak = true;
      }
    }
    if (!has_peak && pure_tel > 0.55f)
    {
      float avg = (vals[0] + vals[1] + vals[2])/3;
      float d1 = abs(vals[0]-vals[1]);
      float d2 = abs(vals[0]-vals[2]);
      float d3 = abs(vals[1]-vals[2]);
      float d_max = max(d1, max(d2, d3));
      float v_max = max(vals[0], max(vals[1], vals[2]));
      log->print(4, "%d:  Max Diff %0.6f",
                 num, d_max/avg);
      if (d_max < FLOOR*avg && v_max < T1_MAX) {
        last_set = 0;
        log->print(4, "  adjusting floor");      
        thres1 = max(thres1, v_max + d_max * FLOOR_ADJ);
        thres2 = max(thres2, thres1*T_MIN_RATIO);
      }
    }
    if (pure_tel < 0.72f && last_set > MAX_LAST_SET) 
    {
      log->print(4, "Reseting Thres");
      thres1 = T1_DEF;
      thres2 = T2_DEF;
      last_set = -1;
    }

    if (thres1 > T1_MAX) thres1 = T1_MAX;
    if (thres2 > T2_MAX) thres2 = T2_MAX;
    
    if (thres1 != old_thres1) {
      if (thres1 != T1_MAX)
        log->print(3, "%d: Thres1 now %#g", num, thres1);
      else
        log->print(3, "%d: T1 Maxed Out at %#g", num, thres1);
    }
    if (thres2 != old_thres2) {
      if (thres2 != T2_MAX)
        log->print(3, "%d: Thres2 now %#g", num, thres2);
      else
        log->print(3, "%d: T2 Maxed Out at %#g", num, thres2);
    }
    old_thres1 = thres1; old_thres2 = thres2;
    
    if (last_set >= 0)
      last_set ++;
    int what;
    if (vals[0] < thres1)       what = 0;
    else if (vals[0] < thres2)  what = 1;
    else                        what = 2;
    FieldDiff res(num-1, vals[0], what);
    vals[0] = vals[1];
    vals[1] = vals[2];

    log->print(2, "Diff %d: %d %#g", res.num, res.what, res.diff);
    return res;
  }
}

//////////////////////////////////////////////////////////////////////
//
// Find Best Frame
//

bool Filter::chose_next_frame()
{
  prev_src_fn = cur_src_fn;
  int ideal_n = prev_src_fn*denm + accum_n + ratio_n;
  int     cur = ideal_n/denm;
  assert(cur < num_src_frames);
  cur_src_fn  = cur;
  int p = prev_src_fn - cur;  // note p does not nessary have to equal -1
  int frac_n  = ideal_n - cur*denm;
  enum Chosen {Either, Cur, Next};
  int chosen  = Cur;

  if (cur == num_src_frames - 1)                          chosen = Cur;
  else if (same_group(0, 1))                              chosen = Either;
  else if (same_group(p, 0))                              chosen = Next;
  else if (similar_group(p,0) && similar_group(-1,0)) {
    if (similar_group(0, 1) && !same_group(1, 2))         chosen = Either;
    else if (diff_group(0, 1))                            chosen = Next;
  } else if (diff_group(p, 0) && diff_group(-1, 0)
             && diff_group(1,2))                          chosen = Either;

  if (chosen == Either) {
    cur_src_fn = frac_n < round_div_n ? cur : cur + 1;
  } else if (chosen == Cur) {
    if (round_div_n < frac_n) round_div_n = frac_n + 1;
    cur_src_fn = cur;
  } else if (chosen == Next) {
    if (round_div_n > frac_n) round_div_n = frac_n - 1;
    cur_src_fn = cur + 1;
  }

  accum_n = ideal_n - cur_src_fn*denm;
  return true;
}

void Filter::get_frame_info()
{
  int matching = is_even(cur_src_fn) ? 1 : -1;
  int other    = -matching;

  enum UseF {Matching, Other, Bob} use_f = Bob;
  
  if      (same_group(0, matching))   use_f = Matching;
  else if (same_group(0, other))      use_f = Other;
  else if (pure_tel > 0.65) {
    if (similar_group(0, matching))   use_f = Matching;
    else if (similar_group(0, other)) use_f = Other;
  }

  RenderInfo & rinf = render_cache[cur_fn];
  rinf.num = cur_fn;
  rinf.parity = get_src_parity(cur_src_fn);
  
  if (use_f == Bob) {
    rinf.bob(cur_src_fn);
  } else if (field_order == FO_TFF) {
    if (is_even(cur_src_fn)) rinf.weave(cur_src_fn, 0, use_f == Matching ? 1 : -1);
    else                     rinf.weave(cur_src_fn, use_f == Matching ? -1 : 1, 0);
  } else {
    if (is_even(cur_src_fn)) rinf.weave(cur_src_fn, use_f == Matching ? 1 : -1, 0);
    else                     rinf.weave(cur_src_fn, 0, use_f == Matching ? -1 : 1);
  }
}

//////////////////////////////////////////////////////////////////////
//
// Render
//

AVS_VideoFrame * Filter::render(const RenderInfo & rinf)
{
  AVS_VideoFrame * res = 0;
  if (rinf.even != rinf.odd) {
    log.print(1, "FRAME %d = [%d,%d]", rinf.num, rinf.even, rinf.odd);
    res = avs_get_frame(weave_src, min(rinf.even, rinf.odd));
  } else {
    log.print(1, "FRAME %d = BOB %d", rinf.num, rinf.even);
    res = avs_get_frame(bob_src, rinf.even);
  }
  return res;
}

//////////////////////////////////////////////////////////////////////
//
// Misc Definations
//

void alloc_clog()
{
  if (gclog) return;
  AllocConsole();
  HANDLE h = GetStdHandle( STD_OUTPUT_HANDLE );
  int fd = _open_osfhandle( (long) h, _O_TEXT );
  gclog = fdopen(fd, "w");
}

void free_clog(void *, AVS_ScriptEnvironment *)
{
  if (!gclog) return;
  fclose(gclog);
  gclog = 0;
  FreeConsole();
}

void Log::use_clog()
{
  alloc_clog();
  clog = gclog;
}

bool Log::open_log_file(const char * fn)
{
  if (flog) return true;
  flog = fopen(fn, "a");
  if (!flog) return false;
  return true;
}

void Log::print(int l, const char * fmt, ...)
{
  if (l > log_level) return;
  char buf[80];
  va_list vl;
  va_start(vl, fmt);
  int n = vsnprintf(buf, 79, fmt, vl);
  if (n < 0 || n > 79) n = 79;
  buf[n] = '\n'; buf[n+1] = '\0';
  va_end(vl);
  if (clog) {
    fputs(buf, clog);
    fflush(clog);}
  if (flog) {
    fputs(buf, flog);}
  if (debug_p)
    OutputDebugString(buf);
}

// this is used by assembler code to benchmark
extern "C" REGPARM void print_cs(unsigned int num, CompareParms * l)
{
  l->log->print(1, "Clock Cycles %d %f", num, (double)num/l->s_width);
}
